@@ -33,24 +33,30 @@ class HUMAP(object):
 			* ANNOY (Python instalation required)
 			* FLANN (Python instalation required)
 
+
 	init (str): (optional, default 'Spectral')
 		Initialization method for the low dimensional embedding. Options include:	
 			* Spectral
 			* random
 
+	reproducible (bool): (optional, default 'False')
+		If the results among different runs need to be reproducible. It affects the runtime execution.
+
 	verbose (bool): (optional, default True)
 		Controls logging.
 
 	"""
-	def __init__(self, levels=np.array([0.2, 0.2]), n_neighbors=100, min_dist=0.15, knn_algorithm='NNDescent', init="Spectral", verbose=True):
+	def __init__(self, levels=np.array([0.2, 0.2]), n_neighbors=100, min_dist=0.15, knn_algorithm='NNDescent', init="Spectral", verbose=True, reproducible=False):
 		self.levels = levels
+		self.n_levels = len(levels)+1
 		self.n_neighbors = n_neighbors
 		self.min_dist = min_dist
 		self.knn_algorithm = knn_algorithm
 		self.verbose = verbose
 		self.init = init
+		self.reproducible = reproducible
 
-		self.h_umap = _hierarchical_umap.HUMAP('precomputed', self.levels, self.n_neighbors, self.min_dist, self.knn_algorithm, self.init, self.verbose)
+		self.h_umap = _hierarchical_umap.HUMAP('precomputed', self.levels, self.n_neighbors, self.min_dist, self.knn_algorithm, self.init, self.verbose, self.reproducible)
 
 
 	def fit(self, X, y=None):
@@ -216,7 +222,7 @@ class HUMAP(object):
 		np.array: the labels for the data points in the specified level
 		"""
 
-		if level <= 0 or level > self.n_levels:
+		if level <= 0 or level >= self.n_levels:
 			raise ValueError("level must be in [1, n_levels-1]")
 		else:			
 			return self.h_umap.get_labels(level)
@@ -262,6 +268,13 @@ class HUMAP(object):
 		self.h_umap.set_fixing_term(fixing_term)
 
 
+	def set_info_file(self, info_file=""):
+		self.h_umap.set_info_file(info_file)
+
+	def set_n_epochs(self, epochs):
+		self.h_umap.set_n_epochs(epochs)
+
+
 	def influence(self, level):
 		r"""
 		Gets the information on how each landmark influence on the subsequent level
@@ -334,8 +347,8 @@ class UMAP(HUMAP):
 		Controls logging.
 
 	"""
-	def __init__(self, n_neighbors=100, min_dist=0.15, knn_algorithm='NNDescent', init="Spectral", verbose=True):
-		super().__init__(np.array([]), n_neighbors, min_dist, knn_algorithm, init, verbose)
+	def __init__(self, n_neighbors=100, min_dist=0.15, knn_algorithm='NNDescent', init="Spectral", verbose=True, reproducible=False):
+		super().__init__(np.array([]), n_neighbors, min_dist, knn_algorithm, init, verbose, reproducible)
 
 	def fit_transform(self, X):
 		"""
